/* -*- Mode: java; tab-width: 8; c-basic-offset: 4 -*-
 * Copyright  1997, 1998 Netscape Communications Corporation, All Rights Reserved.
 */

package com.netscape.javascript.CORBA;

import com.netscape.javascript.*;
import org.omg.CORBA.InterfaceDefPackage.*;
import org.omg.CORBA.AttributeDescription;
import org.omg.CORBA.OperationDescription;
import org.omg.CORBA.InterfaceDef;
import org.omg.CORBA.ParameterDescription;
import org.omg.CORBA.CompletionStatus;
import org.omg.CORBA.ParameterMode;
import org.omg.CORBA.Request;
import org.omg.CORBA.TCKind;
import org.omg.CORBA.Any;

/**
 * This class reflects a CORBA object into the JS
 * environment.
 *
 * @author Mike Shaver
 */
public class CORBAObject extends ScriptableObject {

    public static Object js_Object (Context cx, Object[] args,
				     Function funObj,
				     boolean inNewExpr)
    {
    	if (!inNewExpr)
	    return funObj.construct(args);

/*
        // XXX we should wrap JS objects, perhaps?
    	// XXX Also, permit URL naming and the like
        if (args.length < 1)
                return new CORBAObject();
*/
        // ruslan: why would we allow such case anyway?
        if (args.length < 1 || args.length > 3)
  	    throw new JavaScriptException ("CORBA Object initialization failure: correct syntax is CORBA.Object (repID [,objId] [,host])");

	String repId = ScriptRuntime.toString (args[0]);

    	String objId = null;    // object doesn't need necesserely to have a name
        String  host = null;

    	if (args.length == 2)
            objId = ScriptRuntime.toString(args[1]);
        else
        if (args.length == 3)
            host  = ScriptRuntime.toString(args[2]);

        org.omg.CORBA.Object obj =  Init.orb.bind (repId, objId, host, null);
        return new CORBAObject (obj);
    }

    public String getClassName ()
    {
        return "Object";
    }

    public CORBAObject ()
    {
    	orb = Init.orb; // ruslan: just follow what it was here
    	rep = Init.rep;
    }

    public CORBAObject(org.omg.CORBA.Object obj)
    {
    	this.obj = obj;
    	orb = Init.orb;
    	rep = Init.rep;

	    // OPT: lazily reflect, to allow the no-IR case.
    	reflect ();
    }

    public Object getDefaultValue(Class hint) {
	if (hint == null ||
	    hint == String.class)
	    return "[object CORBAObject: " + repId + "/" + objId + "]";
	return this;
    }

    public Object get (String id, Scriptable start)
    {
        Object obAttr = super.get (id, start);
        if (!(obAttr instanceof AttributeDescription))
                return obAttr;
        AttributeDescription attr = (AttributeDescription)obAttr;

	// OPT: cache getter?
	Request req = obj._request("_get_" + id);
	req.set_return_type(attr.type);
	req.invoke();
	return Marshal.convertFromAny(req.result().value());
    }

    public void put (String id, Scriptable start, Object value) {

        for (int i = 0; i < helper_func.length; i++)
            if (id == helper_func[i])   {
        	super.put (id, start, value);
        	return;
            }

        Object obAttr = super.get (id, start);
        if (!(obAttr instanceof AttributeDescription))
                return;
        AttributeDescription attr = (AttributeDescription)obAttr;

	Request req = obj._request("_set_" + id);
	Any arg = req.add_in_arg();
	arg.type(attr.type);
	if (Marshal.convertIntoAny(value, arg))
	        req.invoke();
    }

    void reflect ()
    {
        repId = obj._repository_id();
        objId = obj._object_name  ();

	// OPT: check repository cache (when we have one)
	InterfaceDef idef =
	    org.omg.CORBA.InterfaceDefHelper.narrow(rep.lookup_id(repId));
	FullInterfaceDescription fulldef =
	    idef.describe_interface();
	for (int i = 0; i < fulldef.attributes.length; i++) {
	    AttributeDescription attr = fulldef.attributes[i];
	    super.put (attr.name.intern(), this, attr);
	    if (attr.mode == org.omg.CORBA.AttributeMode.ATTR_READONLY)
		super.setAttributes(attr.name.intern(), this,
				    ScriptableObject.READONLY | ScriptableObject.PERMANENT);
	    else
		super.setAttributes (attr.name.intern(), this,
				    ScriptableObject.PERMANENT);
	}

	for (int i = 0; i < fulldef.operations.length; i++) {
	    super.put(fulldef.operations[i].name.intern(), this,
		      new CORBAOperationHolder(fulldef.operations[i]));
	    super.setAttributes (fulldef.operations[i].name.intern(), this,
				(short)(ScriptableObject.READONLY | ScriptableObject.PERMANENT));
	}
    }

    static Object invoke(Request req, ParameterDescription[] params,
			 Object[] args) {

	if (args.length != params.length)
	    throw new JavaScriptException("CORBA operation " + req.operation() + " wants " +
					  params.length + " args, only gave " + args.length);

	/*
	 * Check to make sure the args make sense, assigning them as we go.
	 *
	 * I go to moderate lengths below to avoid needing another array for holding the InOut Anys
	 * for later reference.  The code path is as follows:
	 *
	 * for each arg:
	 *   IN: add_in_arg, type, insert value
	 *   OUT: add_out_arg, type, store resulting Any in InOut.value for later reference
	 *   INOUT: add_inout_arg, type, insert value, store resulting Any in InOut
	 *
	 * invoke operation
	 * OPT: (if we had out/inout params)
	 * walk the args looking for InOuts
	 *   when find an InOut, extract value from Any in .value and store result as .value
	 * wrap and return result
	 */

	argWalk: for (int i = 0; i < args.length; i++) {
	    Any any;
	    switch(params[i].mode.value()) {
	    case ParameterMode._PARAM_OUT:
		if (args[i].getClass() != InOut.class)
		    throw new JavaScriptException("CORBA operation " + req.operation() +
						  " has (out " + params[i].name +
						  "), but arg passed wasn't an InOut: " +
						  args[i].getClass().getName());
		any = req.add_out_arg();
		any.type(params[i].type);
		((InOut)args[i]).value = any; // store it in the in/out so that we can get the result later.
		continue argWalk; // We don't need to put any value in, so we just skip ahead
	    case ParameterMode._PARAM_IN:
		any = req.add_in_arg();
		break;
	    case ParameterMode._PARAM_INOUT:
		if (args[i].getClass() != InOut.class)
		    throw new JavaScriptException("CORBA operation " + req.operation() +
						  " has (inout " + params[i].name +
						  "), but arg passed wasn't an InOut: " +
						  args[i].getClass().getName());
		any = req.add_inout_arg();
		break;
	    default:
		throw new JavaScriptException("CORBA operation " + req.operation() +
					      " param " + params[i].name +
					      " has illegal parameter mode: " + params[i].mode);
	    }
	    Object val = args[i];
	    if (params[i].mode.value() == ParameterMode._PARAM_INOUT)
		val = ((InOut)args[i]).value;

	    any.type(params[i].type);
	    if (!Marshal.convertIntoAny (val, any))
		return null;

	    if (params[i].mode == ParameterMode.PARAM_INOUT)
		((InOut)args[i]).value = any; // store it in the in/out so that we can get the result later.
	}

	req.invoke();
	// Now walk the args list again, if we had any (in)out params
	// OPT: if (outParam)
	for (int i = 0; i < args.length; i++)
	    if (params[i].mode == ParameterMode.PARAM_INOUT ||
		params[i].mode == ParameterMode.PARAM_OUT)
		((InOut)args[i]).value = Marshal.convertFromAny((Any)(((InOut)args[i]).value));

        if (req.env ().exception () != null)
                Init.raise (new SystemException ("UNKNOWN", "org.omg.CORBA.UnknownUserException has been thrown", CompletionStatus.COMPLETED_MAYBE));

	return Marshal.convertFromAny (req.return_value());
    }

    org.omg.CORBA.Object obj;
    String repId;
    String objId;

    static org.omg.CORBA.ORB orb;
    static org.omg.CORBA.Repository rep;

    static String [] helper_func = { "constructor" };
}

class CORBAOperationHolder implements Function {
    CORBAOperationHolder(OperationDescription op) {
	this.op = op;
	// OPT: lazily reflect
	type = op.result.kind();
    }

    public Object call(Scriptable thisObj, Object[] args) {
	if (! (thisObj instanceof CORBAObject) ) {
	    throw new com.netscape.javascript.
		JavaScriptException("Attempting to invoke CORBA operation on non-CORBA object.");
	}
	Request req = ((CORBAObject)thisObj).obj._request(op.name);
	req.set_return_type(op.result);
	return CORBAObject.invoke(req, op.parameters, args);
    }

    TCKind type;
    OperationDescription op;

    /* the rest is stuff required by Function (and Scriptable) */

    public Scriptable create() {
	// XXX How do we signal `construction not meaningful'?
	return null;
    }

    public String getClassName() { return "CORBAOperationHolder"; }
    public boolean has(String id, Scriptable start) { return false; }
    public boolean has(int index, Scriptable start) { return false; }
    public Object get (String id, Scriptable start)
    {
        return Scriptable.NOT_FOUND;
    }
    public Object get (int index, Scriptable start)
    {
        return Scriptable.NOT_FOUND;
    }
    public void put (String id, Scriptable start, Object value) { }
    public void put (int index, Scriptable start, Object value) { }
    public short getAttributes(String id, int index, Scriptable start) { return 0; }
    public void setAttributes(String id, int index, Scriptable start, short value) { }
    public void delete (String id) { }
    public void delete (int index) { }
    public Scriptable getPrototype() { return null; }
    public void setPrototype(Scriptable proto) { }
    public Scriptable getParentScope() { return null; }
    public void setParentScope(Scriptable parent) { }
    public Object[] getIds() { return null; }
    public Object getDefaultValue(Class hint) {
	if (hint == null ||
	    hint == String.class)
	    return "[object CORBAOperationHolder: " + (op == null ? "<undefined>" :
						       op.name) + "]";
	return this;
    }

    public Scriptable construct(Object[] args) {
	// XXX How do we signal `construction not meaningful'?
	return null;
    }
}
